from PolyominoTASAssembly import *
from PolyominoTASTileType import *


MODEL_ABSTRACT = 0
MODEL_KINETIC  = 1

class PolyominoSystem:
    def __init__(self):
        self.Clear()
        self.binding_threshold = 2
        self.model = MODEL_ABSTRACT
        self.b_use_rotation = True
        
    def Clear(self):
        self.binding_threshold = None
        self.PolyominoTypeBase_dict = dict()
        self.PolyominoType_dict = dict()
        self.seed = PolyominoAssembly(self)
        self.assembly = PolyominoAssembly(self)
        self.concentrations_dict = dict()
        self.concentrations_total = 0
        self.Gmc = None
        self.Gse = None
        self.max_polyomino_binding = 0
        self.stochastic_event_set = None

    def AddPolyominoType(self, pct):
        if pct.name in self.PolyominoType_dict:
            print(f"PolyominoType with name {pct.name} already exists in set")
        else:
            self.PolyominoTypeBase_dict[pct.name] = pct
            rot = pct.Copy()
            rot.base_type = pct
            self.PolyominoType_dict[rot.name] = rot
            if self.b_use_rotation:
                rotPolys = self.GetRotatedPolyominoes(pct)
                #print(f'Rotated polyominoes: {rotPolys}')
                for rp in rotPolys:
                    rp.base_type = pct
                    self.PolyominoType_dict[rp.name] = rp

            if not pct.concentration is None:
                self.concentrations_total = self.concentrations_total + pct.concentration
                if pct.concentration in self.concentrations_dict:
                    self.concentrations_dict[pct.concentration].append(pct.name)
                else:
                    self.concentrations_dict[pct.concentration] = [pct.name]
            max_binding = pct.GetMaxBinding()
            self.max_polyomino_binding = max(self.max_polyomino_binding, max_binding)

    # This gets copies rotated around the z axis in 90 degree increments, counterclockwise
    def GetRotatedPolyominoes(self, pct):
        rot_polys = []
        for angle in [90,180,270]:
            pctR = pct.GetRotatedVersion(angle)
            rot_polys.append(pctR)
        return rot_polys

    def ReadFromFile(self, file_name, b_viewing_only=False):
        self.Clear()
        xml_tree = ET.parse(file_name)
        xml_root = xml_tree.getroot()
        dim_restrictions = None
        min_binding = None
        if not xml_root.tag == 'PolyominoSystem':
            print("Invalid XML PolyominoSystem file")
            return False
        else:
            seed_polyominoes = []
            assembly_polyominoes = []
            dim_restrictions = (None, None, None)
            for child in xml_root:
                if child.tag == 'bindingThreshold':
                    self.binding_threshold = int(child.text)
                elif child.tag == 'Gmc':
                    self.Gmc = float(child.text)
                elif child.tag == 'Gse':
                    self.Gse = float(child.text)
                elif child.tag == 'rotation' and child.text == 'False':
                    self.b_use_rotation = False
                elif child.tag == 'dim_restrictions':
                    dim_restrictions = eval(child.text)
                elif child.tag == 'min_binding':
                    min_binding = int(child.text)
                elif child.tag == 'PolyominoTypes':
                    for elt_type in child:
                        pct = PolyominoType(name=None, xml_node=elt_type, dim_restrictions = dim_restrictions)
                        self.AddPolyominoType(pct)
                    SimUtils.globalApp.OutputMessage(f'Loaded {len(self.PolyominoType_dict)} polyomino types')
                elif child.tag == 'seed':
                    for child2 in child:
                        if child2.tag == 'Polyominoes':
                            for elt_polyomino in child2:
                                p = Polyomino(polyominoType=None, xml_node=elt_polyomino, polyominoTypes=self.PolyominoType_dict)
                                seed_polyominoes.append(p)
                    SimUtils.globalApp.OutputMessage(f'Loaded {len(seed_polyominoes)} polyominoes in the seed')
                elif child.tag == 'assembly':
                    for child2 in child:
                        if child2.tag == 'Polyominoes':
                            for elt_polyomino in child2:
                                p = Polyomino(polyominoType=None, xml_node=elt_polyomino, polyominoTypes=self.PolyominoType_dict)
                                assembly_polyominoes.append(p)
                    SimUtils.globalApp.OutputMessage(f'Loaded {len(assembly_polyominoes)} polyominoes in the assembly')
        
        # If Gmc and Gse were set, we're using the kinetic model
        if (not self.Gmc is None) and (not self.Gse is None):
            self.model = MODEL_KINETIC
            self.stochastic_event_set = StochasticEventSet(self.max_polyomino_binding, self.concentrations_dict.keys(), self.Gse, self.Gmc)
            SimUtils.globalApp.OutputMessage("Simulation model set to 'kinetic'")
        else:
            self.model = MODEL_ABSTRACT
            SimUtils.globalApp.OutputMessage("Simulation model set to 'abstract'")
            
        if not dim_restrictions is None:
            SimUtils.globalApp.OutputMessage(f"Using dimension restrictions {dim_restrictions}")
            
        self.assembly.polyominoSystem = self
        self.seed.polyominoSystem = self

        if not min_binding is None:
            self.assembly.min_binding = min_binding
            self.seed.min_binding = min_binding

        if len(seed_polyominoes) > 0:
            print(f'Adding {len(seed_polyominoes)} polyominoes in the seed...')
            for p in seed_polyominoes:
                if b_viewing_only:
                    self.seed.AddPolyominoForViewingOnly(p)
                elif self.model == MODEL_KINETIC:
                    self.seed.AddPolyominoKineticModel(p, b_seed=True)
                else:
                    self.seed.AddPolyomino(p)
            SimUtils.globalApp.OutputMessage(f'Added {len(seed_polyominoes)} polyominoes to the seed assembly')
                    
        if len(assembly_polyominoes) > 0:
            print(f'Adding {len(assembly_polyominoes)} polyominoes in the assembly...')
            for p in assembly_polyominoes:
                if b_viewing_only:
                    self.assembly.AddPolyominoForViewingOnly(p)
                elif self.model == MODEL_KINETIC:
                    self.assembly.AddPolyominoKineticModel(p)
                else:
                    self.assembly.AddPolyomino(p)
            SimUtils.globalApp.OutputMessage(f'Added {len(assembly_polyominoes)} polyominoes to the assembly')

        if (len(self.assembly.locationMap) == 0) and (len(self.seed.locationMap) > 0):
            self.assembly = self.seed

        self.assembly.setDimensions()

        # Validation steps:
        # - if given types and seed, verify that each entry in the seed is of one of the types
        # - if given types and assembly, verify that each entry in the assembly is of one of the types
        for loc in self.seed.locationMap:
            polyomino = self.seed.locationMap[loc]
            if not polyomino.polyominoType.name in self.PolyominoType_dict:
                print(f"ERROR: seed contains Polyomino of type {polyomino.polyominoType.name} which is not in the PolyominoType set {self.PolyominoType_dict.keys()}")
                return False
        for loc in self.assembly.locationMap:
            polyomino = self.assembly.locationMap[loc]
            if not polyomino.polyominoType.name in self.PolyominoType_dict:
                print(f"ERROR: assembly contains Polyomino of type {polyomino.polyominoType.name} which is not in the PolyominoType set {self.PolyominoType_dict.keys()}")
                return False
        return True   
                        
    def WriteToFile(self, file_name):
        xml_root = ET.Element('PolyominoSystem')
        if not self.binding_threshold is None:
            elt_bt = ET.SubElement(xml_root, 'bindingThreshold')
            elt_bt.text = str(self.binding_threshold)
        if not self.Gmc is None:
            elt_gmc = ET.SubElement(xml_root, 'Gmc')
            elt_gmc.text = str(self.Gmc)
        if not self.Gse is None:
            elt_gse = ET.SubElement(xml_root, 'Gse')
            elt_gse.text = str(self.Gse)
        if not self.PolyominoType_dict is None:
            elt_pct_set = ET.SubElement(xml_root, 'PolyominoTypes')
            for pct in self.PolyominoType_dict:
                self.PolyominoType_dict[pct].toXML(elt_pct_set)
        if not self.seed is None:
            elt_seed = ET.SubElement(xml_root, 'seed')
            elt_polyominoes = ET.SubElement(elt_seed, 'Polyominoes')
            polyominoes = self.seed.GetPolyominoes()
            for polyomino in polyominoes:
                polyomino.toXML(elt_polyominoes)
        if not self.assembly is None:
            elt_assembly = ET.SubElement(xml_root, 'assembly')
            elt_polyominoes = ET.SubElement(elt_assembly, 'Polyominoes')
            polyominoes = self.assembly.GetPolyominoes()
            for polyomino in polyominoes:
                polyomino.toXML(elt_polyominoes)

        xml_string = ET.tostring(xml_root)
        xml_dom = xml.dom.minidom.parseString(xml_string)
        
        with open(file_name, "w") as f: 
            f.write(xml_dom.toprettyxml())
        return True
