import secrets
from pandas._libs import index


STOCHASTIC_EVENT_TYPE_ATTACHMENT = 0
STOCHASTIC_EVENT_TYPE_DETACHMENT = 1

class DetachmentEventBucket:
    def __init__(self, total_bond_strength, Gse):
        self.total_bond_strength = total_bond_strength
        self.Gse = Gse
        self.event_list = []
        self.size = 0
        
    def get_size(self):
        #TODO: compute the width of the range for this bucket using computation of rate
        return len(self.event_list) #TODO: this will be a more complicated computation
    
    # Randomly select an event from the list and then remove it from the list
    def get_event(self):
        if len(self.event_list) == 0:
            print("ERROR: no attachment events to select from")
            return None
        else:
            index = secrets.randbelow(len(self.event_list))
            event = self.event_list[index]
            del self.event_list[index]
            return event
            
    def add_event(self, event):
        old_size = self.size
        self.event_list.append(event)
        self.size = self.get_size()
        return self.size - old_size
        
    def remove_event(self, event):
        try:
            index = self.event_list.index(event)
            old_size = self.size
            self.event_list.pop(index)
            self.size = self.get_size()
            return self.size - old_size
        except:
            print("ERROR: failed to remove_event - it wasn't found")
            return self.size

class AttachmentEventBucket:
    def __init__(self, concentration, Gmc):
        self.concentration = concentration
        self.num_types = 0
        self.Gmc = Gmc
        self.event_list = []
        self.size = 0
        
    def get_size(self):
        #TODO: compute the size of the bucket for this bucket using computation of rate
        return len(self.event_list) #TODO: this will be a more complicated computation
    
    # Randomly select an event from the list and then remove it from the list
    def get_event(self):
        if len(self.event_list) == 0:
            print("ERROR: no attachment events to select from")
            return None
        else:
            index = secrets.randbelow(len(self.event_list))
            event = self.event_list[index]
            del self.event_list[index]
            return event
            
    def add_event(self, event):
        old_size = self.size
        self.event_list.append(event)
        self.size = self.get_size()
        return self.size - old_size
    
    def remove_event(self, event):
        try:
            index = self.event_list.index(event)
            old_size = self.size
            self.event_list.pop(index)
            self.size = self.get_size()
            return self.size - old_size
        except:
            print("ERROR: failed to remove_event - it wasn't found")
            return self.size
        
class StochasticEvent:
    def __init__(self, type, coords, polyomino=None, bond_strength=None):
        self.type = type
        self.coords = coords
        self.polyomino = polyomino
        
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.type == other.type and self.coords == other.coords:
                if (self.polyomino.polyominoType == None) or (self.polyomino.polyominoType == None):
                    if (self.polyomino.polyominoType == None) and (self.polyomino.polyominoType == None):
                        return True
                    else:
                        return False
                elif self.polyomino.polyominoType.name == other.polyomino.polyominoType.name:
                    return True
        return False
        
class StochasticEventSet:
    def __init__(self, max_bond_strength, concentration_list, Gse, Gmc):
        # Data structures for keeping track of locations where attachment events
        # have been added
        self.attachable_polyominoes = {}
        self.attachment_event_map = {}

        self.range_max = 0

        self.attachment_event_buckets = []
        self.concentration_map = {}
        for concentration in concentration_list:
            if not concentration in self.concentration_map:
                self.attachment_event_buckets.append(AttachmentEventBucket(concentration, Gmc))
                self.concentration_map[concentration] = self.attachment_event_buckets[-1]
                
        self.detachment_event_buckets = []
        for strength in range(max_bond_strength+1):
            self.detachment_event_buckets.append(DetachmentEventBucket(strength, Gse))

        self.event_buckets_list = self.attachment_event_buckets + self.detachment_event_buckets
    
    def select_event(self):
        #TODO: choose a random number in the overall range, then an event from the associated list
        range_select_loc = randinrange(0, self.range_max)

        range_current_loc = 0
        for bucket_num in range(len(self.event_buckets_list)):
            events_bucket = self.event_buckets_list[bucket_num]
            range_current_loc += events_bucket.get_size()
            if range_current_loc >= range_select_loc:
                selected_bucket = bucket_num
                break
        event = events_bucket.get_event()
        if bucket_num < len(self.attachment_event_buckets):
            # it's an attachment event
            self.remove_attachment_event(event)
        else:
            #it's a detachment event
            self.remove_detachment_event(event)
        return event
        
    def add_attachment_event(self, coords, polyomino):
        polyomino_hash = hash(polyomino)
        if polyomino_hash in self.attachable_polyominoes:
            #print("ERROR: in add_attachment_event, polyomino already exists in attachable_polyominoes")
            return
        self.attachable_polyominoes[polyomino_hash] = polyomino
        for block in polyomino:
            if block in self.attachment_event_map:
                if not polyomino_hash in self.attachment_event_map[block]:
                    self.attachment_event_map[block].add(polyomino_hash)
            else:
                self.attachment_event_map[block] = {polyomino_hash} # a set of hashes

        event = StochasticEvent(type=STOCHASTIC_EVENT_TYPE_ATTACHMENT, coords=coords, polyomino=polyomino)
        diff = self.concentration_map[polyomino.polyominoType.concentration].add_event(event)
        self.range_max += diff
        
    def remove_attachment_event(self, event):
        coords = event.coords
        if coords in self.attachment_event_map:
            for polyomino_hash in self.attachment_event_map[coords]:
                polyomino = self.attachable_polyominoes[polyomino_hash]
                for block in polyomino:
                    if not coords == block:
                        self.attachment_event_map[block].remove(polyomino_hash)
                        if len(self.attachment_event_map[block]) == 0:
                            del self.attachment_event_map[block]
                del self.attachable_polyominoes[polyomino_hash]
            self.attachment_event_map[coords].remove(polyomino_hash)
            if len(self.attachment_event_map[coords]) == 0:
                del self.attachment_event_map[coords]
        else:
            print('ERROR: in remove_attachment_event, trying to remove a nonexistent event')
            
        orig_size = self.concentration_map[event.polyomino.polyominoType.concentration].get_size()
        self.concentration_map[event.polyomino.polyominoType.concentration].remove_event(event)     

    def add_detachment_event(self, coords, bond_strength):
        event = StochasticEvent(type=STOCHASTIC_EVENT_TYPE_DETACHMENT, coords=coords, bond_strength=bond_strength)
        diff = self.detachment_event_types[strength].add_event(event)
        self.range_max += diff
        
    def remove_detachment_event(self, event):
        #TODO: fix this, it will not be efficient
        orig_size = self.detachment_event_types[event.strength].get_size()
        self.detachment_event_types[event.strength].event_list.remove(event)
        new_size = self.detachment_event_types[event.strength].get_size()
        self.range_max += (orig_size - new_size)

        
    
        
        
        