from math import exp, log
from collections import deque
import random

from PolyominoTASTileType import Polyomino

class StochasticEvent:
    def __init__(self, coords):
        self.coords = coords
        
class AttachmentEvent(StochasticEvent):
    def __init__(self, coords, polyomino):
        StochasticEvent.__init__(self, coords)
        self.polyomino = polyomino

    def __hash__(self):
        if self.polyomino.polyominoType is None:
            return hash((self.coords, None))
        else:
            return hash((self.coords, self.polyomino.polyominoType.name))

    def __eq__(self, other):
        return self.coords == other.coords and self.polyomino == other.polyomino
        
    def __str__(self):
        return f'AttachmentEvent for coords {self.coords} with polyomino {self.polyomino}'

class DetachmentEvent(StochasticEvent):
    def __init__(self, coords, bond_strength):
        StochasticEvent.__init__(self, coords)
        self.bond_strength = bond_strength

    def __hash__(self):
        return hash((self.coords, self.bond_strength))

    def __eq__(self, other):
        return self.coords == other.coords and self.bond_strength == other.bond_strength
    
    def __str__(self):
        return f'DetachmentEvent for coords {self.coords} with bond_strength {self.bond_strength}'
    
###################################################################################

class StochasticEventBucket:
    def __init__(self):
        self.parent_node = None
        self.events = []
        self.event_indices = {}

    def sample_event(self, remove=True):
        event = random.choice(self.events)
        if remove:
            self.remove_event(event)
        return event

    def add_event(self, event):
        if event not in self.event_indices:
            self.event_indices[event] = len(self.events)
            self.events.append(event)
            return True
        else:
            return False

    def remove_event(self, event):
        if event in self.event_indices:
            index = self.event_indices[event]
            # if we're removing the last element, nothing else needs to be done
            if index == len(self.events) - 1:
                del self.events[index]
                del self.event_indices[event]
            else: # otherwise we'll swap the last element into the deleted position
                last_event = self.events.pop()

                self.events[index] = last_event
                self.event_indices[last_event] = index
                del self.event_indices[event]
            return True
        else:
            return False

class AttachmentEventBucket(StochasticEventBucket):
    def __init__(self, concentration, Gmc):
        StochasticEventBucket.__init__(self)
        self.concentration = concentration
        self.Gmc = Gmc

    def get_log_probability(self):
        if len(self.events) == 0:
            return -1000000
        else:
            return log(len(self.events) * self.concentration) - self.Gmc
        
    def get_probability(self):
        return len(self.events) * self.concentration * exp(-self.Gmc)
    
    def __str__(self):
        return f'AttachmentEventBucket for concentration {self.concentration} with size {len(self.events)}'
        
class DetachmentEventBucket(StochasticEventBucket):
    def __init__(self, strength, Gse):
        StochasticEventBucket.__init__(self)
        self.strength = strength
        self.Gse = Gse

    def get_log_probability(self):
        if len(self.events) == 0:
            return -1000000
        else:
            return log(len(self.events)) - self.strength * self.Gse
        
    def get_probability(self):
        return len(self.events) * exp(-self.strength * self.Gse)
    
    def __str__(self):
        return f'DetachmentEventBucket for strength {self.strength} with size {len(self.events)}'

####################################################################################

class StochasticEventSet:
    def __init__(self, max_bond_strength, concentration_list, Gse, Gmc):
        self.attachment_buckets = {}
        self.detachment_buckets = {}
        self.event_buckets = []

        self.attachment_bucket_indices = {}
        self.detachment_bucket_indices = {}

        for concentration in concentration_list:
            bucket = AttachmentEventBucket(concentration, Gmc)
            self.attachment_buckets[concentration] = bucket
            self.attachment_bucket_indices[concentration] = len(self.event_buckets)
            self.event_buckets.append(bucket)
        
        for strength in range(max_bond_strength+1):
            bucket = DetachmentEventBucket(strength, Gse)
            self.detachment_buckets[strength] = bucket
            self.detachment_bucket_indices[strength] = len(self.event_buckets)
            self.event_buckets.append(bucket)

        # this is a tree that at each node holds a tuple (p_left, p_children)
        # p_left is the cumulative probability of the left children
        # p_children is the cumulative probability of all children
        # this allows updates to be recorded in constant time
        # and when a query is done, the tree is updated in a single O(log(N)) update
        # the query is then performed using an O(log(N)) binary search
        self.cumulative_probability_tree = [(0, 0)] * (2 * len(self.event_buckets) - 1)

        # modified indices are bucket list indices not tree indices
        self.modified_indices = []

    def __str__(self):
        ret = f'Stochastic Event Set: {len(self.event_buckets)} buckets\n'
        
        for bucket in self.event_buckets:
            if isinstance(bucket, AttachmentEventBucket):
                ret += '  Attachment Event Bucket: '
            else:
                ret += '  Detachment Event Bucket: '
            ret += f'{len(bucket.events)} events with total probability {bucket.get_probability()}\n'
            
        return ret
    
    def update_cumulative_probabilities(self):
        if len(self.modified_indices) == 0:
            return
        
        self.modified_indices.sort()
        update_queue = deque()

        for bucket_index in self.modified_indices:
            tree_index = bucket_index + len(self.event_buckets) - 1
            parent = (tree_index - 1) // 2
            bucket = self.event_buckets[bucket_index]
            self.cumulative_probability_tree[tree_index] = (0, bucket.get_probability())
            if len(update_queue) == 0 or update_queue[-1] != parent:
                update_queue.append(parent)

        self.modified_indices.clear()
                
        while len(update_queue) > 0:
            index = update_queue.popleft()
            left = self.cumulative_probability_tree[2 * index + 1]
            right = self.cumulative_probability_tree[2 * index + 2]
            self.cumulative_probability_tree[index] = (left[1], left[1] + right[1])
            parent = (index - 1) // 2
            if parent >= 0:
                if len(update_queue) == 0 or update_queue[-1] != parent:
                    update_queue.append(parent)

    def add_attachment_event(self, coords, polyomino):
        concentration = polyomino.polyominoType.concentration
        bucket_index = self.attachment_bucket_indices[concentration]
        event = AttachmentEvent(coords, polyomino)
        if self.event_buckets[bucket_index].add_event(event):
            self.modified_indices.append(bucket_index)
            return True
        else:
            return False

    def add_detachment_event(self, coords, bond_strength):
        bucket_index = self.detachment_bucket_indices[bond_strength]
        event = DetachmentEvent(coords, bond_strength)
        if self.event_buckets[bucket_index].add_event(event):
            self.modified_indices.append(bucket_index)
            return True
        else:
            return False

    def remove_attachment_event(self, coords, polyomino):
        concentration = polyomino.polyominoType.concentration
        bucket_index = self.attachment_bucket_indices[concentration]
        event = AttachmentEvent(coords, polyomino)
        if self.event_buckets[bucket_index].remove_event(event):
            self.modified_indices.append(bucket_index)
            return True
        else:
            return False

    def remove_detachment_event(self, coords, bond_strength):
        bucket_index = self.detachment_bucket_indices[bond_strength]
        event = DetachmentEvent(coords, bond_strength)
        if self.event_buckets[bucket_index].remove_event(event):
            self.modified_indices.append(bucket_index)
            return True
        else:
            return False

    def remove_event(self, event):
        if isinstance(event, AttachmentEvent):
            return self.remove_attachment_event(event.coords, event.polyomino)
        elif isinstance(event, DetachmentEvent):
            return self.remove_detachment_event(event.coords, event.bond_strength)
        else:
            return False

    def sample_event(self, remove=True):
        gumbel_samples = []
        for i, bucket in enumerate(self.event_buckets):
            lp = bucket.get_log_probability()
            u = random.random()
            g = -log(-log(u))
            
            gumbel_samples.append((g + lp, i))

        gumbel_samples.sort(reverse=True)
        
        gi = 0
        bucket = None
        while bucket is None or len(bucket.events) == 0:
            bucket_index = gumbel_samples[gi][1]
            bucket = self.event_buckets[bucket_index]
            gi += 1

        event = bucket.sample_event(False)
        if remove:
            removed = self.remove_event(event)
            if not removed:
                print("ERROR Something went wrong trying to remove sampled event")
                
        return event
        
    
    def sample_event_old(self, remove=True):
        self.update_cumulative_probabilities()

        cumulative_probability = self.cumulative_probability_tree[0][1]
        selection = random.random() * cumulative_probability

        # perform a binary search to
        offset = 0
        tree_index = 0
        found = False

        while tree_index < len(self.event_buckets) - 1:
            left_probability = self.cumulative_probability_tree[tree_index][0]
            
            if selection < offset + left_probability:
                tree_index = 2 * tree_index + 1
            else:
                tree_index = 2 * tree_index + 2
                offset += left_probability

        bucket_index = tree_index - len(self.event_buckets) + 1
        bucket = self.event_buckets[bucket_index]

        if len(bucket.events) == 0:
            print("ERROR found an empty bucket somehow")

        
        event = bucket.sample_event(False)
        #print(f'DEBUG: event = {event} from bucket {bucket}')
        if remove:
            removed = self.remove_event(event)
            if not removed:
                print("ERROR Something went wrong trying to remove the sampled event")
        return event
        
